﻿using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using UnityEngine;

/// <summary>
/// Base class for all Gadgets to derive from.
/// A Gadget is a tool for the player (and level editor) to interact with the GameWorld.
/// </summary>
public abstract class Gadget
{
    /// <summary>Used to map to a T:Gadget </summary>
    /// <remarks>Do NOT rename elements! Do NOT change values! Deserialization relies on it!</remarks>
    public enum GadgetIDs
    {
        Undefined = -1,
        Pointer = 0,
        Tape = 1,
        AngleTool = 2,
        LineTool = 3,
        LotTool = 4,
        Pendulum = 5,
        PoleTool = 6,
        Remover = 7,
    }

    public static Dictionary<Type, GadgetIDs> GadgetTypeToIDs = new(){
        {typeof(UndefinedGadget)    , GadgetIDs.Undefined },
        {typeof(Pointer)            , GadgetIDs.Pointer },
        {typeof(Tape)               , GadgetIDs.Tape },
        {typeof(AngleTool)          , GadgetIDs.AngleTool },
        {typeof(LineTool)           , GadgetIDs.LineTool },
        {typeof(LotTool)            , GadgetIDs.LotTool },
        {typeof(Pendulum)           , GadgetIDs.Pendulum },
        {typeof(PoleTool)           , GadgetIDs.PoleTool },
        {typeof(Remover)            , GadgetIDs.Remover },
        };

    /// <summary> Position in tool belt. </summary>
    /// <remarks>Set in Inspector or <see cref="Awake"/></remarks>
    public int Rank;
    /// <summary> Tool Name </summary>
    /// <remarks>Set in Inspector or <see cref="Awake"/></remarks>
    public string UiName;
    /// <summary> Maximum range for this Tool. For consistency use GadgetDistances in <see cref="GlobalBehaviour"/>.</summary>
    /// <remarks>Set in Inspector or <see cref="Awake"/></remarks>
    public float MaxRange;
    public float MaxHeight;

    /// <summary>Which sprite to use</summary>
    public int ButtonIndx;
    public int MaterialIndx;
    /// <summary>Layers to ignore for this gadget by default.</summary>
    /// <remarks>Set in Inspector</remarks>
    public LayerMask IgnoreLayerMask;
    public LayerMask SecondaryLayerMask;

    //protected List<Vector3> LinePositions = new();

    /// <summary>
    /// Collection of <c>Type</c>s of *all* available <see cref="Gadget"/>s to choose from.
    /// </summary>
    [JsonIgnore]
    public static readonly IEnumerable<Type> GadgetTypes = Assembly.GetExecutingAssembly().GetTypes().Where(t => t != typeof(Gadget) && typeof(Gadget).IsAssignableFrom(t));

    public Gadget()
    {
        Type this_type = this.GetType();
        if (!GadgetTypeToIDs.ContainsKey(this_type)) {
            Debug.LogError("No " + this_type.ToString() + "in Dictionary<Type, GadgetIDs> GadgetTypeToIDs!");
            return;
        }
        var GadgetID = GadgetTypeToIDs[this_type];

        if (!GadgetManager.DataContainerGadgetDict.ContainsKey(GadgetID)) {
            Debug.LogError("No " + GadgetID.ToString() + "in assigned "+ typeof(DataContainerGadgetCollection).Name +"!");
            return;
        }
        var data_cache = GadgetManager.DataContainerGadgetDict[GadgetID];

        Rank = data_cache.Rank;
        UiName = data_cache.UiName;
        MaxRange = data_cache.MaxRange;
        MaxHeight = data_cache.MaxHeight;
        IgnoreLayerMask = data_cache.IgnoreLayerMask;
        SecondaryLayerMask = data_cache.SecondaryLayerMask;
        ButtonIndx = data_cache.ButtonIndx < GadgetManager.ButtonSprites.Length
            ? data_cache.ButtonIndx : 0;
        MaterialIndx = data_cache.MaterialIndx < GadgetManager.Materials.Length
            ? data_cache.MaterialIndx : 0;
    }

    public void Awake()
    {
        _Awake();
    }

    public void Enable()
    {
        GadgetManager.Cursor.setLayerMask(~IgnoreLayerMask.value);
        GadgetManager.Cursor.MaxRange = MaxRange;

        ResetGadget();
        _Enable();
    }

    public void Disable()
    {
        ResetGadget();
        _Disable();
    }

    public void Update()
    {
        if (!CommunicationEvents.GadgetCanBeUsed)
            return;

        if (GadgetManager.LineRenderer.enabled)
            UpdateLineDrawing();

        _Update();
    }

    public void Hit(RaycastHit hit)
    {
        if (!CommunicationEvents.GadgetCanBeUsed
            || hit.transform.position.y > MaxHeight)
            return;

        _Hit(hit);
    }

    public virtual void _Awake() { }
    public virtual void _Enable() { }
    public virtual void _Disable() { }
    public virtual void _Update() { }

    /// <summary>
    /// Called when <see cref="CommunicationEvents.TriggerEvent"/> is invoked, a.k.a. when Player clicks in GameWorld.
    /// </summary>
    /// <param name="hit">the position where it was clicked</param>
    public virtual void _Hit(RaycastHit hit) { }

    protected void ActivateLineDrawing()
    {
        GadgetManager.LineRenderer.enabled = true;
        GadgetManager.LineRenderer.material = GadgetManager.Materials[MaterialIndx];

        _ActivateLineDrawing();
    }

    protected void DeactivateLineDrawing()
    {
        GadgetManager.LineRenderer.positionCount = 0;
        GadgetManager.LineRenderer.enabled = false;

        _DeactivateLineDrawing();
    }

    protected void UpdateLineDrawing()
    {
        _UpdateLineDrawing();
    }

    protected virtual void _ActivateLineDrawing() { }
    protected virtual void _DeactivateLineDrawing() { }
    protected virtual void _UpdateLineDrawing() { }

    protected Vector3 GetPosition(int i)
        => GadgetManager.LineRenderer.GetPosition(i);
    protected void SetPosition(int i, Vector3 v)
        => GadgetManager.LineRenderer.SetPosition(i, v);

    public void ResetGadget()
    {
        DeactivateLineDrawing();
        _ResetGadget();
    }

    protected virtual void _ResetGadget() { }

    public class UndefinedGadget : Gadget { }
}